/*
    mcbootmenu.c
    James Higgs 2003

    based on mc_example.c (libmc API sample).
*/

#include <tamtypes.h>
#include <kernel.h>
#include <sifrpc.h>
#include <loadfile.h>
#include <fileio.h>
#include <stdio.h>
#include <iopcontrol.h>
#include <libmc.h>

#include "hw.h"
#include "hardware.h"

#include "font5200.c"

#include "libpad.h"

//#define DEBUG

#define ROM_PADMAN

#if defined(ROM_PADMAN) && defined(NEW_PADMAN)
#error Only one of ROM_PADMAN & NEW_PADMAN should be defined!
#endif

#if !defined(ROM_PADMAN) && !defined(NEW_PADMAN)
#error ROM_PADMAN or NEW_PADMAN must be defined!
#endif

/*
 * Global var's
 */
// pad_dma_buf is provided by the user, one buf for each pad
// contains the pad's current state
static char padBuf[256] __attribute__((aligned(64)));

static char actAlign[6];
static int actuators;


// ELF-loading stuff
#define ELF_MAGIC	0x464c457f
#define ELF_PT_LOAD	1
typedef struct {
	u8	ident[16];
	u16	type;
	u16	machine;
	u32	version;
	u32	entry;
	u32	phoff;
	u32	shoff;
	u32	flags;
	u16	ehsize;
	u16	phentsize;
	u16	phnum;
	u16	shentsize;
	u16	shnum;
	u16	shstrndx;
} elf_header_t;

typedef struct {
	u32	type;
	u32	offset;
	void	*vaddr;
	u32	paddr;
	u32	filesz;
	u32	memsz;
	u32	flags;
	u32	align;
} elf_pheader_t;

extern void *_end;


#define TYPE_MC
//#define TYPE_XMC


#define WIDTH   366
#define HEIGHT  256
#define FRAMERATE   4
#define STATUS_Y 232

unsigned char *Img;						// pntr to blit buffer
int g_nWhichBuffer = 0;
int show_logo = 0;
char sStatus[128];

// simple print funcs
void drawChar(char c, int x, int y, unsigned int colour);
void printXY(char *s, int x, int y, unsigned int colour);
int do_select_menu(void);


void LoadModules(void);
void LoadAndRunMCElf(char *filename);

#define ARRAY_ENTRIES	64
static mcTable mcDir[ARRAY_ENTRIES] __attribute__((aligned(64)));
static int mc_Type, mc_Free, mc_Format;
int num_mc_files;
unsigned char romver[16];
char mcport = 0;

// clear emulation screen buffer
// * optimise by using 32 or 64 bit writes
void clrEmuScreen(unsigned char colour) {
    unsigned char *pp;
    unsigned int numbytes;

    pp = pScreen;
    numbytes = g_nScreen_X * g_nScreen_Y;
    while(numbytes--) *pp++ = colour;
}

// draw a char using the system font
void drawChar(char c, int x, int y, unsigned int colour) {
    unsigned int i, j;
    unsigned char cc;
    unsigned char *pp;
    unsigned char *pc;

    // set character pointer
    pc = &font5200[(c-32)*8];

    // set screen pointer
    pp = pScreen + x + (g_nScreen_X * y);
    for(i=0; i<8; i++) {
        cc = *pc++;
        for(j=0; j<8; j++) {
            if(cc & 0x80) *pp = colour;
            cc = cc << 1;
            pp++;
        }
        pp += (g_nScreen_X - 8);
    }
}

// draw a string of characters
void printXY(char *s, int x, int y, unsigned int colour) {

    while(*s) {
        drawChar(*s++, x, y, colour);
        x += 8;
    }
}


/** PutImage() ***********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void PutImage(void)
{
    U8 *pScr;
    U8 *pLine;
    int x, y;
/*
    // copy Coleco buffer (Img) to PS2frame screen (pScreen)
	// NB: only copy 256x192 from Img to pScreen
    pScr = pScreen;
    for(y=0; y<HEIGHT; y++) {
        pLine = Img + (y * WIDTH);  // skip top 8 lines
        for(x=0; x<WIDTH; x++) {    // *** OPTMISE ***
            *pScr++ = *pLine++;
        }
    }
*/
	UpdateScreen();		// uploads+and renders new screen. (palette is also updated every frame for effects)

	while (TestVRstart() < FRAMERATE); 					// wait for FRD number of vblanks
	ClearVRcount();
    g_nWhichBuffer ^= 1;
	SetupScreen( g_nWhichBuffer );					// FLIP!!!

}

// Get PS2's ROM version / region
void GetROMVersion(void) {
    int fd = fioOpen("rom0:ROMVER", O_RDONLY);
	fioRead(fd, romver, sizeof romver);
	fioClose(fd);

    romver[15] = 0;
}

void PrintMCFiles(int highlighted) {
    int i;
    char s[128];

    for(i=0; i<num_mc_files; i++) {
		if(mcDir[i].attrFile & MC_ATTR_SUBDIR)
			sprintf(s, " %s (dir)", mcDir[i].name);
		else
			sprintf(s, " %s (%d)", mcDir[i].name, mcDir[i].fileSizeByte);

        if(highlighted == i) s[0] = '*';

        printXY(s, 16, i*8 + 40, 2);
    }
}

void ReadMCDir(void) {
    int ret;
    char MCSystemPath[128] = "/BEDATA-SYSTEM/*";

	MCSystemPath[2] = (romver[4] == 'E' ? 'E' :
			(romver[4] == 'J' ? 'I' : 'A'));

	mcGetDir(mcport, 0, MCSystemPath, 0, ARRAY_ENTRIES - 10, mcDir);
	mcSync(0, NULL, &ret);
#ifdef DEBUG
	printf("mcGetDir returned %d listing system directory %s on memory card %d\n\n", ret, MCSystemPath, mcport);
#endif
    num_mc_files = ret;
}

void ReadHostDir(void) {

}

void ReDraw(int highlighted) {
    int i;
    clrEmuScreen(0);
    printXY("Jum's MC Boot Menu", 4, 4, 1);
    printXY("Select an ELF to run:", 4, 24, 1);
    printXY("+-----------------------------+", 4, 32, 1);
    for(i=40; i<200; i+=8) {
    printXY("|                             |", 4, i, 1);
    }
    printXY("+-----------------------------+", 4, 200, 1);
    PrintMCFiles(highlighted);
    printXY(sStatus, 4, STATUS_Y, 2);
    PutImage();
}

void jprintf(char *s) {
#ifdef DEBUG
    printf("%s\n", s);
#endif
    printXY(s, 4, 232, 2);
}

char *SelectELF(void) {
    int ret;
    int selected = 0;
    int highlighted = 0;
    struct padButtonStatus buttons;
    u32 paddata;
    u32 old_pad = 0;
    u32 new_pad;

    while(!selected) {
        ret = padRead(0, 0, &buttons); // port, slot, buttons
            
        if (ret != 0) {
            paddata = 0xffff ^ ((buttons.btns[0] << 8) | 
                                buttons.btns[1]);
                
            new_pad = paddata & ~old_pad;
            old_pad = paddata;
                
            // Directions
            if(new_pad & PAD_LEFT) ;
            if(new_pad & PAD_DOWN) {
                if(highlighted < num_mc_files-1) highlighted++;
            }
            if(new_pad & PAD_RIGHT) ;
            if(new_pad & PAD_UP) {
                if(0 < highlighted) highlighted--;
            }
            if((new_pad & PAD_START) || (new_pad & PAD_CROSS)) {
                sprintf(sStatus, "-  %s -", mcDir[highlighted].name);
                selected = 1;
            }
        }
//        sprintf(sStatus, "%X", new_pad);
        ReDraw(highlighted);
    }

    return mcDir[highlighted].name;
}

// Init MC stuff and get which port the MC is plugged into
// - returns 0 if there is a card in port 0
// - else returns 1 if there is a card in port 1
// - else returns -11
int InitMC(void) {
    int ret;

#ifdef TYPE_MC
	if(mcInit(MC_TYPE_MC) < 0) {
		jprintf("Failed to initialise memcard server!\n");
		SleepThread();
	}
#else
	if(mcInit(MC_TYPE_XMC) < 0) {
		jprintf("Failed to initialise memcard server!\n");
		SleepThread();
	}
#endif

#ifdef DEBUG
    printf("mcInit() OK!");
#endif

	// int mcGetInfo(int port, int slot, int* type, int* free, int* format);
	//
	// mcGetInfo retrieves memcard state info, such as the memcard type, free blocks and 
	// the format status.
	//
	// mcGetInfo is passed pointers to three variables, which are filled upon completion 
	// of the getinfo rpc command. The mcGetInfo return values are as follows:
	//
	//  0 : The same memory card has been connected since the last mcGetInfo call.
	// -1 : Switched to a formatted memory card.
	// -2 : Switched to an unformatted memory card.
	// -10 or less : The memory card could not be detected.
	//
	// NOTE: With the MCMAN/MCSERV, *format is always returned as 0 regardless of if
	//       if the memcard is formatted or not.

	// Check port 0 for an MC
    // (should return <= -10 if there isn't)
	mcGetInfo(0, 0, &mc_Type, &mc_Free, &mc_Format);
	mcSync(0, NULL, &ret);
#ifdef DEBUG
	printf("mcGetInfo returned %d\n",ret);
	printf("Type: %d Free: %d Format: %d\n\n", mc_Type, mc_Free, mc_Format);
#endif

    if( -1 == ret ) return 0;

    // if there is no MC in port 0, then check port 1
	mcGetInfo(1,0,&mc_Type,&mc_Free,&mc_Format);
	mcSync(0, NULL, &ret);
#ifdef DEBUG
	printf("mcGetInfo returned %d\n",ret);
	printf("Type: %d Free: %d Format: %d\n\n", mc_Type, mc_Free, mc_Format);
#endif

    if( -1 == ret ) return 1;

    return -11;
}

/*
 * waitPadReady()
 */
int waitPadReady(int port, int slot)
{
    int state;
    int lastState;
    char stateString[16];

    state = padGetState(port, slot);
    lastState = -1;
    while((state != PAD_STATE_STABLE) && (state != PAD_STATE_FINDCTP1)) {
        if (state != lastState) {
            padStateInt2String(state, stateString);
//            scr_printf("Please wait, pad(%d,%d) is in state %s\n",
//                       port, slot, stateString);
        }
        lastState = state;
        state=padGetState(port, slot);
    }
    // Were the pad ever 'out of sync'?
    if (lastState != -1) {
        jprintf("Pad OK!\n");
    }
}


/*
 * initializePad()
 */
int initializePad(int port, int slot)
{

    int ret;
    int modes;
    int i;
    char s[128];

    waitPadReady(port, slot);

    // How many different modes can this device operate in?
    // i.e. get # entrys in the modetable
    modes = padInfoMode(port, slot, PAD_MODETABLE, -1);
    sprintf(s, "The device has %d modes", modes);
    jprintf(s);

    if (modes > 0) {
        for (i = 0; i < modes; i++) {
            sprintf(s, "uses mode %d ", padInfoMode(port, slot, PAD_MODETABLE, i));
            jprintf(s);
        }
    }

    sprintf(s, "It is currently using mode %d",
               padInfoMode(port, slot, PAD_MODECURID, 0));
    jprintf(s);

    // If modes == 0, this is not a Dual shock controller 
    // (it has no actuator engines)
    if (modes == 0) {
//        scr_printf("This is a digital controller?\n");
        return 1;
    }

    return 1;
}


// *************************************************************************
// *** MAIN
// *************************************************************************
int main() {

    int fd, ret,ret2;
    int i,j;
    int size;
    char *buffer;
    char *pc;
    char s[256];
    char cargar[50];
    char c;
    char cpath[256];
    int selected = 0;
    int highlighted = 0;
    struct padButtonStatus buttons;
    u32 paddata;
    u32 old_pad = 0;
    u32 new_pad;	
    int cd=0;

	// Initialise
	SifInitRpc(0);
	LoadModules();

    // Initialise platform stuff
	// detect system, and reset GS into that mode
	if(pal_ntsc() == 3) {
		initGraph(3);
	} else {
		initGraph(2);
	}

	g_nWhichBuffer = 0;
	SetupScreen( g_nWhichBuffer );						// set up MY screen!!!
    *GS_PMODE = 0x0066;

    *GS_BGCOLOR = 0x00FF00; // GREEN

    install_VRstart_handler();

    *GS_BGCOLOR = 0x00FFFF; // YELLOW

    // set palette
    for(i=0; i<16; i++) {
        SetPaletteEntry(0x00000000, 0);
        SetPaletteEntry(0x00808080, 1);
        SetPaletteEntry(0x00FFFFFF, 2);
    }

    // *** Required BEFORE calling ReadMCDir etc
    GetROMVersion();
#ifdef DEBUG
    printf("Rom version: %s\n", romver);
#endif

    // Initialise MC stuff and get port which MC is plugged into
    mcport = InitMC();
#ifdef DEBUG
    printf("InitMC() returned %d\n", mcport);
#endif

    // If no MC found, then sleep
    if(mcport < 0) {
        strcpy(sStatus, "No MC found in either port!");
        ReDraw(0);
        SleepThread();
    }

    ReadMCDir();
    strcpy(sStatus, "Ready.\0");
    //ReDraw(0);

    // get the pad going
    padInit(0);
    if((ret = padPortOpen(0, 0, padBuf)) == 0) {
        jprintf("padOpenPort failed");
        SleepThread();
    }

jprintf("padPortOpen() OK\n");

    if(!initializePad(0, 0)) {
        jprintf("pad initalization failed!");
        SleepThread();
    }

jprintf("initializePad() OK\n");

    ret=padGetState(0, 0);
    while((ret != PAD_STATE_STABLE) && (ret != PAD_STATE_FINDCTP1)) {
        if(ret==PAD_STATE_DISCONN) {
            jprintf("Pad is disconnected");
        }
        ret=padGetState(0, 0);
    }

jprintf("Starting SelectELF()...\n");
	ret2 = padRead(0, 0, &buttons); // port, slot, buttons
	while(ret2==0) {ret2 = padRead(0, 0, &buttons); // port, slot, buttons
			printf("Leyendo pad\n");
			}
	paddata = 0xffff ^ ((buttons.btns[0] << 8) | 
                                buttons.btns[1]);
                
        new_pad = paddata & ~old_pad;
        old_pad = paddata;
                
        // Directions
        if(new_pad & PAD_START) cd=1;


	if(cd==0) {	
	printf("Leyendo desde cd\n");
    	if ((fd = fioOpen("cdrom0:\\SYSTEM.CNF", O_RDONLY)) < 0) {
    	} 
	j=0;

	
	c=fioGetc(fd);
	
	while(c!=';') {		
	 c=fioGetc(fd);
 	 j++;
	
	}
	
	j=0;
	c=fioGetc(fd);
	
	while(c!='\\')	{	
	 c=fioGetc(fd);
	 j++;
	}
	
	j=0;
	while(c!=';') {
		c=fioGetc(fd);
		cargar[j]=c;
		j++;
	}	
	cargar[j-1]='\0';
	//tenemos el sles en cargar
	strcpy(cpath,"cdrom0:\\");
	strcat(cpath,cargar);


	//Cargamos juego desde cd
	SifExitRpc();
	LoadExecPS2(cpath, 0, NULL);
	}

	else {
	printf("Leyendo desde MC\n");
	strcpy(s,"mc0:/BEDATA-SYSTEM/MCLOADER.ELF");
	LoadAndRunMCElf(s);
	SleepThread();	   
	}	
	
}


void LoadModules(void)
{
    int ret;

#ifdef TYPE_MC
	ret = SifLoadModule("rom0:SIO2MAN", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: SIO2MAN");
		SleepThread();
	}

	ret = SifLoadModule("rom0:CDVDMAN", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: CDVDMAN");
		SleepThread();
	}

	ret = SifLoadModule("rom0:MCMAN", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: MCMAN");
		SleepThread();
	}

	ret = SifLoadModule("rom0:MCSERV", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: MCSERV");
		SleepThread();
	}
#else
	ret = SifLoadModule("rom0:XSIO2MAN", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: SIO2MAN");
		SleepThread();
	}

	ret = SifLoadModule("rom0:XMCMAN", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: MCMAN");
		SleepThread();
	}

	ret = SifLoadModule("rom0:XMCSERV", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: MCSERV");
		SleepThread();
	}
	
	ret = SifLoadModule("rom0:CDVDMAN", 0, NULL);
	if (ret < 0) {
		jprintf("Failed to load module: CDVDMAN");
		SleepThread();
	}


#endif

#ifdef ROM_PADMAN
    ret = SifLoadModule("rom0:PADMAN", 0, NULL);
#else
    ret = SifLoadModule("rom0:XPADMAN", 0, NULL);
#endif 
    if (ret < 0) {
        jprintf("Failed to load PAD module");
        SleepThread();
    }
}


void LoadAndRunMCElf(char *filename) {
	u8 *boot_elf = (u8 *)&_end;
	elf_header_t *eh = (elf_header_t *)boot_elf;
	elf_pheader_t *eph;
	char *argv[1];
	void *pdata;
	int fd, size, i;

	if ((fd = fioOpen(filename, O_RDONLY)) < 0) {
        *GS_BGCOLOR = 0xFF0000; // RED
		goto error;
	}

	size = fioLseek(fd, 0, SEEK_END);
	if (!size) {
		fioClose(fd);
        *GS_BGCOLOR = 0x00FF00; // GREEN
		goto error;
	}

	fioLseek(fd, 0, SEEK_SET);

	fioRead(fd, boot_elf, size);
	fioClose(fd);

	/* Load the ELF into RAM.  */
	if (_lw((u32)&eh->ident) != ELF_MAGIC) {
        *GS_BGCOLOR = 0x0000FF; // BLUE
		goto error;
	}

	eph = (elf_pheader_t *)(boot_elf + eh->phoff);

	/* Scan through the ELF's program headers and copy them into RAM, then
	   zero out any non-loaded regions.  */
	for (i = 0; i < eh->phnum; i++) {
		if (eph[i].type != ELF_PT_LOAD)
			continue;

		pdata = (void *)(boot_elf + eph[i].offset);
		memcpy(eph[i].vaddr, pdata, eph[i].filesz);

		if (eph[i].memsz > eph[i].filesz)
			memset(eph[i].vaddr + eph[i].filesz, 0,
					eph[i].memsz - eph[i].filesz);
	}

	/* Let's go.  */
	argv[0] = filename;

	fioExit();
	SifResetIop();

	FlushCache(0);
	FlushCache(2);

	ExecPS2((void *)eh->entry, 0, 1, argv);


error:
	while (1) ;

}
